@@ -1,5 +1,3 @@
-context("testing lgb.Dataset functionality")
-
 data(agaricus.train, package = "lightgbm")
 train_data <- agaricus.train$data[seq_len(1000L), ]
 train_label <- agaricus.train$label[seq_len(1000L)]
@@ -10,38 +8,32 @@ test_label <- agaricus.test$label[1L:100L]
 
 test_that("lgb.Dataset: basic construction, saving, loading", {
   # from sparse matrix
-  dtest1 <- lgb.Dataset(test_data, label = test_label)
+  dtest1 <- lgb.Dataset(
+    test_data
+    , label = test_label
+    , params = list(
+      verbose = .LGB_VERBOSITY
+    )
+  )
   # from dense matrix
   dtest2 <- lgb.Dataset(as.matrix(test_data), label = test_label)
-  expect_equal(getinfo(dtest1, "label"), getinfo(dtest2, "label"))
   expect_equal(get_field(dtest1, "label"), get_field(dtest2, "label"))
 
   # save to a local file
   tmp_file <- tempfile("lgb.Dataset_")
   lgb.Dataset.save(dtest1, tmp_file)
   # read from a local file
-  dtest3 <- lgb.Dataset(tmp_file)
+  dtest3 <- lgb.Dataset(
+    tmp_file
+    , params = list(
+      verbose = .LGB_VERBOSITY
+    )
+  )
   lgb.Dataset.construct(dtest3)
   unlink(tmp_file)
-  expect_equal(getinfo(dtest1, "label"), getinfo(dtest3, "label"))
   expect_equal(get_field(dtest1, "label"), get_field(dtest3, "label"))
 })
 
-test_that("lgb.Dataset: getinfo & setinfo", {
-  dtest <- lgb.Dataset(test_data)
-  dtest$construct()
-
-  setinfo(dtest, "label", test_label)
-  labels <- getinfo(dtest, "label")
-  expect_equal(test_label, getinfo(dtest, "label"))
-
-  expect_true(length(getinfo(dtest, "weight")) == 0L)
-  expect_true(length(getinfo(dtest, "init_score")) == 0L)
-
-  # any other label should error
-  expect_error(setinfo(dtest, "asdf", test_label))
-})
-
 test_that("lgb.Dataset: get_field & set_field", {
   dtest <- lgb.Dataset(test_data)
   dtest$construct()
@@ -67,34 +59,6 @@ test_that("lgb.Dataset: slice, dim", {
   expect_equal(ncol(dsub1), ncol(test_data))
 })
 
-test_that("Dataset$slice() supports passing additional parameters through '...'", {
-  dtest <- lgb.Dataset(test_data, label = test_label)
-  dtest$construct()
-  dsub1 <- slice(
-    dataset = dtest
-    , idxset = seq_len(42L)
-    , feature_pre_filter = FALSE
-  )
-  dsub1$construct()
-  expect_identical(dtest$get_params(), list())
-  expect_identical(dsub1$get_params(), list(feature_pre_filter = FALSE))
-})
-
-test_that("Dataset$slice() supports passing Dataset attributes through '...'", {
-  dtest <- lgb.Dataset(test_data, label = test_label)
-  dtest$construct()
-  num_subset_rows <- 51L
-  init_score <- rnorm(n = num_subset_rows)
-  dsub1 <- slice(
-    dataset = dtest
-    , idxset = seq_len(num_subset_rows)
-    , init_score = init_score
-  )
-  dsub1$construct()
-  expect_null(dtest$getinfo("init_score"), NULL)
-  expect_identical(dsub1$getinfo("init_score"), init_score)
-})
-
 test_that("Dataset$set_reference() on a constructed Dataset fails if raw data has been freed", {
   dtrain <- lgb.Dataset(train_data, label = train_label)
   dtrain$construct()
@@ -165,7 +129,7 @@ test_that("Dataset$set_reference() updates categorical_feature, colnames, and pr
   dtrain$construct()
   bst <- Booster$new(
     train_set = dtrain
-    , params = list(verbose = -1L)
+    , params = list(verbose = -1L, num_threads = .LGB_MAX_THREADS)
   )
   dtrain$.__enclos_env__$private$predictor <- bst$to_predictor()
 
@@ -189,7 +153,10 @@ test_that("Dataset$set_reference() updates categorical_feature, colnames, and pr
   dtest$set_reference(dtrain)
 
   # after setting reference to dtrain, those attributes should have dtrain's values
-  expect_is(dtest$.__enclos_env__$private$predictor, "lgb.Predictor")
+  expect_true(methods::is(
+    dtest$.__enclos_env__$private$predictor
+    , "lgb.Predictor"
+  ))
   expect_identical(
     dtest$.__enclos_env__$private$predictor$.__enclos_env__$private$handle
     , dtrain$.__enclos_env__$private$predictor$.__enclos_env__$private$handle
@@ -216,7 +183,9 @@ test_that("lgb.Dataset: colnames", {
     colnames(dtest) <- "asdf"
   })
   new_names <- make.names(seq_len(ncol(test_data)))
-  expect_silent(colnames(dtest) <- new_names)
+  expect_silent({
+    colnames(dtest) <- new_names
+  })
   expect_equal(colnames(dtest), new_names)
 })
 
@@ -237,10 +206,10 @@ test_that("lgb.Dataset: Dataset should be able to construct from matrix and retu
     , rawData
     , nrow(rawData)
     , ncol(rawData)
-    , lightgbm:::lgb.params2str(params = list())
+    , lightgbm:::.params2str(params = list())
     , ref_handle
   )
-  expect_is(handle, "externalptr")
+  expect_true(methods::is(handle, "externalptr"))
   expect_false(is.null(handle))
   .Call(LGBM_DatasetFree_R, handle)
   handle <- NULL
@@ -259,23 +228,12 @@ test_that("cpp errors should be raised as proper R errors", {
     , init_score = seq_len(10L)
   )
   expect_error({
-    dtrain$construct()
+    capture.output({
+      dtrain$construct()
+    }, type = "message")
   }, regexp = "Initial score size doesn't match data size")
 })
 
-test_that("lgb.Dataset$setinfo() should convert 'group' to integer", {
-  ds <- lgb.Dataset(
-    data = matrix(rnorm(100L), nrow = 50L, ncol = 2L)
-    , label = sample(c(0L, 1L), size = 50L, replace = TRUE)
-  )
-  ds$construct()
-  current_group <- ds$getinfo("group")
-  expect_null(current_group)
-  group_as_numeric <- rep(25.0, 2L)
-  ds$setinfo("group", group_as_numeric)
-  expect_identical(ds$getinfo("group"), as.integer(group_as_numeric))
-})
-
 test_that("lgb.Dataset$set_field() should convert 'group' to integer", {
   ds <- lgb.Dataset(
     data = matrix(rnorm(100L), nrow = 50L, ncol = 2L)
@@ -364,7 +322,7 @@ test_that("Dataset$update_parameters() does nothing for empty inputs", {
   res <- ds$update_params(
     params = list()
   )
-  expect_true(lgb.is.Dataset(res))
+  expect_true(.is_Dataset(res))
 
   new_params <- ds$get_params()
   expect_identical(new_params, initial_params)
@@ -385,7 +343,7 @@ test_that("Dataset$update_params() works correctly for recognized Dataset parame
   res <- ds$update_params(
     params = new_params
   )
-  expect_true(lgb.is.Dataset(res))
+  expect_true(.is_Dataset(res))
 
   updated_params <- ds$get_params()
   for (param_name in names(new_params)) {
@@ -398,23 +356,26 @@ test_that("Dataset$finalize() should not fail on an already-finalized Dataset",
     data = test_data
     , label = test_label
   )
-  expect_true(lgb.is.null.handle(dtest$.__enclos_env__$private$handle))
+  expect_true(.is_null_handle(dtest$.__enclos_env__$private$handle))
 
   dtest$construct()
-  expect_false(lgb.is.null.handle(dtest$.__enclos_env__$private$handle))
+  expect_false(.is_null_handle(dtest$.__enclos_env__$private$handle))
 
   dtest$finalize()
-  expect_true(lgb.is.null.handle(dtest$.__enclos_env__$private$handle))
+  expect_true(.is_null_handle(dtest$.__enclos_env__$private$handle))
 
   # calling finalize() a second time shouldn't cause any issues
   dtest$finalize()
-  expect_true(lgb.is.null.handle(dtest$.__enclos_env__$private$handle))
+  expect_true(.is_null_handle(dtest$.__enclos_env__$private$handle))
 })
 
 test_that("lgb.Dataset: should be able to run lgb.train() immediately after using lgb.Dataset() on a file", {
   dtest <- lgb.Dataset(
     data = test_data
     , label = test_label
+    , params = list(
+      verbose = .LGB_VERBOSITY
+    )
   )
   tmp_file <- tempfile(pattern = "lgb.Dataset_")
   lgb.Dataset.save(
@@ -430,6 +391,8 @@ test_that("lgb.Dataset: should be able to run lgb.train() immediately after usin
     , metric = "binary_logloss"
     , num_leaves = 5L
     , learning_rate = 1.0
+    , verbose = .LGB_VERBOSITY
+    , num_threads = .LGB_MAX_THREADS
   )
 
   # should be able to train right away
@@ -438,13 +401,16 @@ test_that("lgb.Dataset: should be able to run lgb.train() immediately after usin
     , data = dtest_read_in
   )
 
-  expect_true(lgb.is.Booster(x = bst))
+  expect_true(.is_Booster(x = bst))
 })
 
 test_that("lgb.Dataset: should be able to run lgb.cv() immediately after using lgb.Dataset() on a file", {
   dtest <- lgb.Dataset(
     data = test_data
     , label = test_label
+    , params = list(
+      verbosity = .LGB_VERBOSITY
+    )
   )
   tmp_file <- tempfile(pattern = "lgb.Dataset_")
   lgb.Dataset.save(
@@ -460,6 +426,9 @@ test_that("lgb.Dataset: should be able to run lgb.cv() immediately after using l
     , metric = "binary_logloss"
     , num_leaves = 5L
     , learning_rate = 1.0
+    , num_iterations = 5L
+    , verbosity = .LGB_VERBOSITY
+    , num_threads = .LGB_MAX_THREADS
   )
 
   # should be able to train right away
@@ -468,14 +437,14 @@ test_that("lgb.Dataset: should be able to run lgb.cv() immediately after using l
     , data = dtest_read_in
   )
 
-  expect_is(bst, "lgb.CVBooster")
+  expect_true(methods::is(bst, "lgb.CVBooster"))
 })
 
 test_that("lgb.Dataset: should be able to use and retrieve long feature names", {
   # set one feature to a value longer than the default buffer size used
   # in LGBM_DatasetGetFeatureNames_R
   feature_names <- names(iris)
-  long_name <- paste0(rep("a", 1000L), collapse = "")
+  long_name <- strrep("a", 1000L)
   feature_names[1L] <- long_name
   names(iris) <- feature_names
   # check that feature name survived the trip from R to C++ and back
@@ -502,7 +471,10 @@ test_that("lgb.Dataset: should be able to create a Dataset from a text file with
 
   dtrain <- lgb.Dataset(
     data = train_file
-    , params = list(header = TRUE)
+    , params = list(
+      header = TRUE
+      , verbosity = .LGB_VERBOSITY
+    )
   )
   dtrain$construct()
   expect_identical(dtrain$get_colnames(), c("x1", "x2"))
@@ -523,7 +495,10 @@ test_that("lgb.Dataset: should be able to create a Dataset from a text file with
 
   dtrain <- lgb.Dataset(
     data = train_file
-    , params = list(header = FALSE)
+    , params = list(
+      header = FALSE
+      , verbosity = .LGB_VERBOSITY
+    )
   )
   dtrain$construct()
   expect_identical(dtrain$get_colnames(), c("Column_0", "Column_1"))
@@ -554,6 +529,9 @@ test_that("Dataset: method calls on a Dataset with a null handle should raise an
   expect_error({
     dtrain$get_colnames()
   }, regexp = "cannot get column names before dataset has been constructed")
+  expect_error({
+    dtrain$get_feature_num_bin(1L)
+  }, regexp = "Cannot get number of bins in feature before constructing Dataset.")
   expect_error({
     dtrain$save_binary(fname = tempfile(fileext = ".bin"))
   }, regexp = "Attempting to create a Dataset without any raw data")
@@ -578,3 +556,68 @@ test_that("Dataset: method calls on a Dataset with a null handle should raise an
     dtrain$set_reference(reference = dvalid)
   }, regexp = "cannot get column names before dataset has been constructed")
 })
+
+test_that("lgb.Dataset$get_feature_num_bin() works", {
+  raw_df <- data.frame(
+    all_random = runif(100L)
+    , two_vals = rep(c(1.0, 2.0), 50L)
+    , three_vals = c(rep(c(0.0, 1.0, 2.0), 33L), 0.0)
+    , two_vals_plus_missing = c(rep(c(1.0, 2.0), 49L), NA_real_, NA_real_)
+    , all_zero = rep(0.0, 100L)
+    , categorical = sample.int(2L, 100L, replace = TRUE)
+  )
+  n_features <- ncol(raw_df)
+  raw_mat <- data.matrix(raw_df)
+  min_data_in_bin <- 2L
+  ds <- lgb.Dataset(
+    raw_mat
+    , params = list(min_data_in_bin = min_data_in_bin)
+    , categorical_feature = n_features
+  )
+  ds$construct()
+  expected_num_bins <- c(
+    100L %/% min_data_in_bin + 1L  # extra bin for zero
+    , 3L  # 0, 1, 2
+    , 3L  # 0, 1, 2
+    , 4L  # 0, 1, 2 + NA
+    , 0L  # unused
+    , 3L  # 1, 2 + NA
+  )
+  actual_num_bins <- sapply(1L:n_features, ds$get_feature_num_bin)
+  expect_identical(actual_num_bins, expected_num_bins)
+  # test using defined feature names
+  bins_by_name <- sapply(colnames(raw_mat), ds$get_feature_num_bin)
+  expect_identical(unname(bins_by_name), expected_num_bins)
+  # test using default feature names
+  no_names_mat <- raw_mat
+  colnames(no_names_mat) <- NULL
+  ds_no_names <- lgb.Dataset(
+    no_names_mat
+    , params = list(min_data_in_bin = min_data_in_bin)
+    , categorical_feature = n_features
+  )
+  ds_no_names$construct()
+  default_names <- lapply(
+    X = seq(1L, ncol(raw_mat))
+    , FUN = function(i) {
+      sprintf("Column_%d", i - 1L)
+    }
+  )
+  bins_by_default_name <- sapply(default_names, ds_no_names$get_feature_num_bin)
+  expect_identical(bins_by_default_name, expected_num_bins)
+})
+
+test_that("lgb.Dataset can be constructed with categorical features and without colnames", {
+  # check that dataset can be constructed
+  raw_mat <- matrix(rep(c(0L, 1L), 50L), ncol = 1L)
+  ds <- lgb.Dataset(raw_mat, categorical_feature = 1L)$construct()
+  sparse_mat <- as(raw_mat, "dgCMatrix")
+  ds2 <- lgb.Dataset(sparse_mat, categorical_feature = 1L)$construct()
+  # check that the column names are the default ones
+  expect_equal(ds$.__enclos_env__$private$colnames, "Column_0")
+  expect_equal(ds2$.__enclos_env__$private$colnames, "Column_0")
+  # check for error when index is greater than the number of columns
+  expect_error({
+    lgb.Dataset(raw_mat, categorical_feature = 2L)$construct()
+  }, regexp = "supplied a too large value in categorical_feature: 2 but only 1 features")
+})
