@@ -1,5 +1,3 @@
-context(".METRICS_HIGHER_BETTER()")
-
 test_that(".METRICS_HIGHER_BETTER() should be well formed", {
     metrics <- .METRICS_HIGHER_BETTER()
     metric_names <- names(.METRICS_HIGHER_BETTER())
@@ -8,5 +6,5 @@ test_that(".METRICS_HIGHER_BETTER() should be well formed", {
     # no metrics should be repeated
     expect_true(length(unique(metric_names)) == length(metrics))
     # should not be any NAs
-    expect_false(any(is.na(metrics)))
+    expect_false(anyNA(metrics))
 })
