@@ -1,5 +1,3 @@
-context("Case weights are respected")
-
 test_that("Gamma regression reacts on 'weight'", {
   n <- 100L
   set.seed(87L)
@@ -7,7 +5,7 @@ test_that("Gamma regression reacts on 'weight'", {
   y <- X[, 1L] + X[, 2L] + runif(n)
   X_pred <- X[1L:5L, ]
 
-  params <- list(objective = "gamma")
+  params <- list(objective = "gamma", num_threads = .LGB_MAX_THREADS)
 
   # Unweighted
   dtrain <- lgb.Dataset(X, label = y)
@@ -15,7 +13,7 @@ test_that("Gamma regression reacts on 'weight'", {
     params = params
     , data = dtrain
     , nrounds = 4L
-    , verbose = 0L
+    , verbose = .LGB_VERBOSITY
   )
   pred_unweighted <- predict(bst, X_pred)
 
@@ -29,7 +27,7 @@ test_that("Gamma regression reacts on 'weight'", {
     params = params
     , data = dtrain
     , nrounds = 4L
-    , verbose = 0L
+    , verbose = .LGB_VERBOSITY
   )
   pred_weighted_1 <- predict(bst, X_pred)
 
@@ -43,7 +41,7 @@ test_that("Gamma regression reacts on 'weight'", {
     params = params
     , data = dtrain
     , nrounds = 4L
-    , verbose = 0L
+    , verbose = .LGB_VERBOSITY
   )
   pred_weighted_2 <- predict(bst, X_pred)
 
@@ -57,7 +55,7 @@ test_that("Gamma regression reacts on 'weight'", {
     params = params
     , data = dtrain
     , nrounds = 4L
-    , verbose = 0L
+    , verbose = .LGB_VERBOSITY
   )
   pred_weighted <- predict(bst, X_pred)
 
